/* Indentation. */
div.doc-contents:not(.first) {
  padding-left: 25px;
  border-left: .05rem solid var(--md-typeset-table-color);
}

/* Avoid breaking parameters name, etc. in table cells. */
td code {
  word-break: normal !important;
}
